SCHWAB_API_BASE_URL = "https://api.schwabapi.com/v1"
SCHWAB_MARKET_DATA_API_BASE_URL = "https://api.schwabapi.com/marketdata/v1"
SCHWAB_TRADER_API_BASE_URL = "https://api.schwabapi.com/trader/v1"
SCHWAB_AUTH_PATH = "oauth/authorize"
SCHWAB_TOKEN_PATH = "oauth/token"
